"""
Enumerating spatial graph diagrams
==================================

Current code is limited to trivalent system architectures.

The basic approach differs somewhat from the one in the paper.
Namely, I use "plantri" to enumerate possible diagram shadows with the
specified number of crossings::

  http://users.cecs.anu.edu.au/~bdm/plantri/

You need to compile plantri and have it in the same directory as this
file (or somewhere in your path) for the enumeration to work.

Due to a limitation of plantri, this restricts us to shadows which are
"diagrammatically prime" in that there is not a circle meeting the
shadow in two points that has vertices of the shadow on both sides.
Equivalently, the dual planar graph is simple.

If the system architecture graph cannot be disconnected by removing
two edges, this only excludes shadows all of whose spatial diagrams
are the connect sum of a spatial graph diagram with the desired system
architecture and a knot.  Presumably, we would want to exclude such in
any event.  However, the example in Case Study 1 can be so
disconnected...

Validation
==========

Compared to Dobrynin and Vesnin:

1. For the theta graph, the list of Yamada polynomials through 5
   crossings matches after removing the non-prime examples from their
   list (theta_3, theta_5, theta_10, theta_14).

2. For the tetrahedral graph, the list of Yamada polynomials through 4
   crossings matches after removing the non-prime Omega_5.

Note: The way this script is written w/ pickling you must import this script into another script
rather than directly calculate Yamada polynomials in this script (you'll get error messages)

TODO Make Reidemeister Class
R0 -> if two adjacent crossings on the same edge have different orientation then simplify
R1 -> Switch crossing order
R2 -> if two edges share two crossings then check if you can simplify them
R3 -> You can slide an edge
R4 -> if one edge is fully under or over a set then move...
R5 -> You can untwist a bunch
R6 -> ?


"""

import networkx as nx
import pickle
from cypari import pari
from .H_polynomial import h_poly
from .utilities import get_coefficients_and_exponents




class EntryPoint:
    def __init__(self, vertex_like, index):
        self.vertex = vertex_like
        self.index = index

    def __eq__(self, other):
        return self.vertex == other.vertex and self.index == other.index

    def __hash__(self):
        return hash((self.vertex, self.index))

    def __repr__(self):
        return "<EP %s %d>" % (self.vertex, self.index)

    def next_corner(self):
        """
        Moves around face clockwise.
        """
        V, i = self.vertex, self.index
        j = (i + 1) % V.degree
        W, k = V.adjacent[j]
        return EntryPoint(W, k)


class BaseVertex:
    """
    A flat vertex has n inputs, labeled 0, 1, ..., n-1 in
    anticlockwise order.

    The adjacents should be edges, not other flat vertices.
    """

    def __init__(self, degree, label):
        self.label = label
        self.degree = degree
        self.adjacent = degree * [None]

    def __getitem__(self, i):
        return self, i % self.degree

    def __setitem__(self, i, other):
        o, j = other
        i = i % self.degree
        self.adjacent[i] = other
        o.adjacent[j] = (self, i)

    def __repr__(self):
        return repr(self.label)

    def entry_points(self):
        return [EntryPoint(self, i) for i in range(self.degree)]

    def next_available_index(self):
        """
        Returns the next available index.

        This is useful for automatically generating then setting items.

        TODO Verify indexing is correct, since degree should start with 1 not 0
        """
        for i in range(self.degree):
            if self.adjacent[i] is None:
                return i
        raise ValueError("No available index")

    def already_assigned(self, node):
        """
        Returns true if the index is already assigned
        """

        already_assigned = []

        for adjacent in self.adjacent:
            if adjacent is None:
                already_assigned.append(False)
            else:
                if adjacent[0] == node:
                    already_assigned.append(True)
                else:
                    already_assigned.append(False)

        return any(already_assigned)


class Vertex(BaseVertex):
    pass


class Crossing(BaseVertex):
    """
    A crossing has four inputs, labeled 0, 1, 2, 3 in anticlockwise
    order.  Convention is 0 - 2 crosses *under* 1, 3.
    """

    def __init__(self, label):
        BaseVertex.__init__(self, 4, label)

    def flow(self, i):
        return self.adjacent[(i + 2) % 4]


class Edge(BaseVertex):
    def __init__(self, label):
        BaseVertex.__init__(self, 2, label)

    def fuse(self):
        """
        Joins the incoming and outgoing strands and removes
        self from the picture.
        """
        (A, i), (B, j) = self.adjacent
        A[i] = B[j]

    def flow(self, i):
        return self.adjacent[(i + 1) % 2]


class Reidemeister:

    """
    Reidemeister moves zero through six

    TODO Implement has_r0, has_r3, has_r4, has_r5
    """

    def __init__(self):
        self.edges = None
        self.crossings = None
        self.vertices = None

    def has_r1(self):
        for C in self.crossings:
            for i in range(4):
                E, e = C.adjacent[i]
                D, d = E.flow(e)
                if D == C and (i + 1) % 4 == d:
                    return True
        return False

    def has_r2(self):
        for E in self.edges:
            A, a = E.adjacent[0]
            if isinstance(A, Crossing):
                B, b = E.adjacent[1]
                if isinstance(B, Crossing):
                    if (a + b) % 2 == 0:
                        return True
        return False

    def has_r6(self):
        for V in self.vertices:
            for i in range(V.degree):
                E, e = V.adjacent[i]
                A, a = E.flow(e)
                if isinstance(A, Crossing):
                    E, e = V.adjacent[(i + 1) % V.degree]
                    B, b = E.flow(e)
                    if A == B and (b + 1) % 4 == a:
                        return True
        return False






class SpatialGraphDiagram(Reidemeister):
    """

    """

    def __init__(self, data, check=True):

        # Initialize Reidemeister class
        Reidemeister.__init__(self)

        # Need labels of vertices/crossings to be unique and hashable
        self.data = {d.label: d for d in data}
        assert len(data) == len(self.data)
        self.crossings = [d for d in data if isinstance(d, Crossing)]
        self.vertices = [d for d in data if isinstance(d, Vertex)]
        self.edges = [d for d in data if isinstance(d, Edge)]

        self._merge_vertices()

        if len(self.edges) == 0 and len(data) > 0:
            self._inflate_edges()



        if check:
            self._check()

    def faces(self):
        """
        The faces are the complementary regions of the diagram. Each
        face is given as a list of corners of BaseVertices as one goes
        around *clockwise*. These corners are recorded as
        EntryPoints, where EntryPoints(c, j) denotes the corner of the
        face abutting crossing c between strand j and j + 1.

        Alternatively, the sequence of EntryPoints can be regarded as
        the *heads* of the oriented edges of the face.
        """

        entry_points = []

        for V in self.data.values():
            entry_points += V.entry_points()

        corners = set(entry_points)
        faces = []

        while len(corners):
            face = [corners.pop()]
            while True:
                next_ep = face[-1].next_corner()
                if next_ep == face[0]:
                    faces.append(face)
                    break
                else:
                    corners.remove(next_ep)
                    face.append(next_ep)

        return faces

    def euler(self):
        """
        Returns the Euler characteristic of the diagram.
        """
        v = len(self.crossings) + len(self.vertices)
        e = len(self.edges)
        f = len(self.faces())
        return v - e + f

    def is_planar(self):
        """
        Returns True if the diagram is planar.
        """
        return self.euler() == 2 * len(list(nx.connected_components(self.projection_graph())))

    def _check(self):
        """
        Checks that the diagram is valid.
        """

        assert 2 * len(self.edges) == sum(d.degree for d in self.crossings + self.vertices)

        for C in self.crossings:
            assert all(isinstance(v, Edge) for v, j in C.adjacent)
        for V in self.vertices:
            assert all(isinstance(v, Edge) for v, j in V.adjacent)
        for E in self.edges:
            assert all(not isinstance(v, Edge) for v, j in E.adjacent)
        assert self.is_planar()

    def _inflate_edges(self):
        """
        Creates edges from the crossings and vertices.
        """

        curr_index = 0
        edges = []

        for A in self.crossings + self.vertices:
            for i in range(A.degree):
                B, j = A.adjacent[i]
                if not isinstance(B, Edge):
                    E = Edge(curr_index)
                    curr_index += 1
                    edges.append(E)
                    self.data[E.label] = E
                    E[0] = (A, i)
                    E[1] = (B, j)

        self.edges = edges

    def _merge_vertices(self):
        """
        Removes 2-valent vertices from the diagram. These vertices increase the complexity and runtime of
        calculations but do not add any information.
        """

        degree_two = [vertex for vertex in self.vertices if vertex.degree == 2]
        for vertex in degree_two:
            A, i = vertex.adjacent[0]
            B, j = vertex.adjacent[1]
            if A != vertex and B != vertex:
                A[i] = B[j]
                self.vertices.remove(vertex)
                self.data.pop(vertex.label)

    def copy(self):
        """
        Returns a serialized copy of the diagram.
        """
        return pickle.loads(pickle.dumps(self))

    def projection_graph(self):
        """
        TODO Add documentation
        """

        G = nx.MultiGraph()

        for e in self.edges:
            v = e.adjacent[0][0].label
            w = e.adjacent[1][0].label
            G.add_edge(v, w)
        return G

    def underlying_graph(self):
        """
        TODO Add documentation
        """

        G = nx.MultiGraph()
        vertex_inputs = set()

        for V in self.vertices:
            vertex_inputs.update((V, i) for i in range(V.degree))

        edges_used = 0

        while len(vertex_inputs):
            V, i = vertex_inputs.pop()
            W, j = V.adjacent[i]
            while not isinstance(W, Vertex):
                if isinstance(W, Edge):
                    edges_used += 1
                W, j = W.flow(j)
            vertex_inputs.remove((W, j))
            v, w = V.label, W.label
            G.add_edge(v, w)

        if edges_used == len(self.edges):
            return G

    def underlying_planar_embedding(self):
        """
        Returns the underlying planar embedding of an abstract graph.
        """

        G = nx.PlanarEmbedding()
        parts = self.vertices + self.crossings + self.edges

        for A in parts:
            for i in range(A.degree):
                B, j = A.adjacent[i]
                ref_nbr = None if i == 0 else A.adjacent[i - 1][0].label
                G.add_half_edge_ccw(A.label, B.label, ref_nbr)

        G.check_structure()
        return G

    def remove_crossing(self, C):
        """
        Removes a crossing from the diagram.
        """

        self.crossings.remove(C)
        self.data.pop(C.label)

    def remove_edge(self, E):
        """
        Removes an edge from the diagram.
        """

        self.edges.remove(E)
        self.data.pop(E.label)

    def add_vertex(self, V):
        """
        Adds a vertex to the diagram.
        """

        self.vertices.append(V)
        self.data[V.label] = V

    def short_cut(self, crossing, i0):
        """
        Short-cuts a crossing by removing the edge between them.
        """

        i1 = (i0 + 1) % 4
        E0, j0 = crossing.adjacent[i0]
        E1, j1 = crossing.adjacent[i1]
        if E0 == E1:
            V0 = Vertex(2, repr(E0) + '_stopper')
            self.add_vertex(V0)
            V0[0] = E0[j0]
            V0[1] = E1[j1]
        else:
            E1[j1] = E0[j0]
            E1.fuse()
            self.remove_edge(E1)



    def yamada_polynomial(self, check_pieces=False):
        """
        Return the non-normalized Yamada polynomial of the knot.
        """
        A = pari('A')

        if len(self.crossings) == 0:
            return h_poly(self.projection_graph())

        C = self.crossings[0]
        c = C.label

        # S_plus
        S_plus = self.copy()
        C_plus = S_plus.data[c]
        S_plus.remove_crossing(C_plus)
        S_plus.short_cut(C_plus, 0)
        S_plus.short_cut(C_plus, 2)
        if check_pieces:
            S_plus._check()

        # S_minus
        S_minus = self.copy()
        C_minus = S_minus.data[c]
        S_minus.remove_crossing(C_minus)
        S_minus.short_cut(C_minus, 1)
        S_minus.short_cut(C_minus, 3)
        if check_pieces:
            S_minus._check()

        # S_0
        S_0 = self.copy()
        C_0 = S_0.data[c]
        S_0.remove_crossing(C_0)

        V = Vertex(4, repr(C_0) + '_smushed')
        S_0.add_vertex(V)

        for i in range(4):
            B, j = C_0.adjacent[i]
            V[i] = B[j]

        if check_pieces:
            S_0._check()

        Y_plus  = S_plus.yamada_polynomial()
        Y_minus = S_minus.yamada_polynomial()
        Y_0     = S_0.yamada_polynomial()

        return A * Y_plus + (A ** -1) * Y_minus + Y_0

    def normalized_yamada_polynomial(self):
        """normalized_yamada_polynomial

        TODO Why does the normalizer work this way?
        """

        yamada_polynomial = self.yamada_polynomial()

        A = pari('A')

        _, exps = get_coefficients_and_exponents(yamada_polynomial)
        a, b = min(exps), max(exps)
        ans1 = (-A) ** (-a) * yamada_polynomial
        ans2 = (-A) ** b * reverse_poly(yamada_polynomial)

        normalized_yamada_polynomial = min([ans1, ans2], key=list)

        return normalized_yamada_polynomial


def normalize_yamada_polynomial(yamada_polynomial):
    """normalized_yamada_polynomial

    TODO Why does the normalizer work this way?
    """

    A = pari('A')

    _, exps = get_coefficients_and_exponents(yamada_polynomial)
    a, b = min(exps), max(exps)
    ans1 = (-A) ** (-a) * yamada_polynomial
    ans2 = (-A) ** b * reverse_poly(yamada_polynomial)

    normalized_yamada_polynomial = min([ans1, ans2], key=list)

    return normalized_yamada_polynomial





def reverse_poly(poly):
    """
    TODO Why does reverse_poly invert the exponent? What is the purpose?
    """

    A = pari('A')

    coeffs, exps = get_coefficients_and_exponents(poly)

    ans = pari(0)

    for c, e in zip(coeffs, exps):
        ans += c * A ** (-e)

    return ans



import numpy as np
import networkx as nx
import json

def read_json_file(filename):
    """
    Reads a json file and returns a dictionary
    """

    with open(filename) as f:
        data = json.load(f)
    return data

def read_json_files(filenames):
    """
    Reads a list of json files and returns a dictionary
    """

    data = {}
    for filename in filenames:
        data.update(read_json_file(filename))
    return data


def split_edge(edge):
    """
    Splits an edge with more than two nodes into a series of edges which each have two nodes.
    """

    edges = []
    for i in range(len(edge)-1):
        edges.append((edge[i], edge[i+1]))
    return edges


def split_edges(edges):
    """
    Splits a list of edges with more than two nodes into a series of edges which each have two nodes.
    """

    # Strip crossings from edges
    edges_without_crossings = []
    for edge in edges:
        edge_without_crossings = [node for node in edge if "C" not in node]
        edges_without_crossings.append(edge_without_crossings)

    new_edges = []
    for edge in edges_without_crossings:
        new_edges.extend(split_edge(edge))
    return new_edges


def extract_graph_from_json_file(filename):
    """
    Reads a json file and returns a networkx graph
    """

    data = read_json_file(filename)

    # Extract the dictionary inputs
    nodes              = list(data['3D_positions'][0].keys())
    node_positions     = list(data['3D_positions'][0].values())
    crossings          = list(data['3D_positions'][1].keys())
    crossing_positions = list(data['3D_positions'][1].values())
    edges              = list(data['3D_positions'][2])

    # Extract non crossings from crossings
    noncrossings, noncrossing_positions = zip(*[(crossing, crossing_position) for crossing, crossing_position in zip(crossings, crossing_positions) if "C" not in crossing])



    # Only extract non crossings
    # Merge nodes and crossings
    nodes.extend(noncrossings)
    node_positions.extend(noncrossing_positions)

    # Remove crossings from edges

    # Format edges so that each edge has two nodes
    edges = split_edges(edges)

    return nodes, node_positions, edges



# def generate_random_layout(layout):
#     """
#     Generates random layouts using a force-directed algorithm

#     Initially assumes 0 rotation and 1x6 design vector


#     :return:
#     """

#     g = nx.MultiGraph()
#     g.add_nodes_from(layout.nodes)
#     g.add_edges_from(layout.edges)

#     # Optimal distance between nodes
#     k = 1

#     scale = 6

#     # TODO remove this random number seed for actual problems
#     seed = 11

#     # Dimension of layout
#     dim = 3

#     positions = nx.spring_layout(g, k=k, dim=dim, scale=scale, seed=seed)

#     # Generate random angles too?

#     # Temporarily pad zeros for rotations
#     design_vectors = []
#     rotation = np.array([0, 0, 0])

#     for i in positions:
#         position = positions[i]
#         design_vector = np.concatenate((position, rotation))
#         design_vectors.append(design_vector)

#     # Flatten design vectors
#     # TODO Make more efficient?
#     design_vector = np.concatenate(design_vectors)

#     return design_vector


# """Generate Yamada polynomial spatial topologies..."""


# import networkx as nx
# import pickle
# import collections
# from cypari import pari
# import itertools
# import subprocess
# import string
# import io
# import time





# def shadows_via_plantri_by_ascii(num_tri_verts, num_crossings):
#     assert num_tri_verts % 2 == 0
#     vertices = num_tri_verts + num_crossings
#     edges = (3 * num_tri_verts + 4 * num_crossings) // 2
#     faces = 2 - vertices + edges
#     cmd = ['./plantri',
#            '-p -d',  # simple planar maps, but return the dual
#            '-f4',  # maximum valence in the returned dual is <= 4
#            '-c1',  # graph should be 1-connected
#            '-m2',  # no valence 1 vertices = no loops in the dual
#            '-a',  # return in ascii format
#            '-e%d' % edges,
#            '%d' % faces]
#     proc = subprocess.run(' '.join(cmd), shell=True, text=True, capture_output=True)
#     ans = []
#     for line in proc.stdout.splitlines():
#         graph_data = line.split()[1].split(',')
#         counts = collections.Counter(len(v) for v in graph_data)
#         assert counts[3] == num_tri_verts and counts[4] == num_crossings
#         ans.append(graph_data)
#     return ans


# def read_edge_code(stream, size):
#     """
#     Read 1 byte form of edge code
#     """
#     ans = [[]]
#     for _ in range(size):
#         i = int.from_bytes(stream.read(1), 'big')
#         if i < 255:
#             ans[-1].append(i)
#         else:
#             ans.append([])
#     return ans


# def shadows_via_plantri_by_edge_codes(num_tri_verts, num_crossings):
#     assert num_tri_verts % 2 == 0
#     vertices = num_tri_verts + num_crossings
#     edges = (3 * num_tri_verts + 4 * num_crossings) // 2
#     faces = 2 - vertices + edges
#     cmd = ['./plantri',
#            '-p -d',  # simple planar maps, but return the dual
#            '-f4',  # maximum valence in the returned dual is <= 4
#            '-c1',  # graph should be 1-connected
#            '-m2',  # no valence 1 vertices = no loops in the dual
#            '-E',  # return binary edge code format
#            '-e%d' % edges,
#            '%d' % faces]
#     proc = subprocess.run(' '.join(cmd), shell=True, capture_output=True)
#     stdout = io.BytesIO(proc.stdout)
#     assert stdout.read(13) == b'>>edge_code<<'
#     shadows = []
#     while True:
#         b = stdout.read(1)
#         if len(b) == 0:
#             break
#         size = int.from_bytes(b, 'big')
#         assert size != 0
#         shadows.append(read_edge_code(stdout, size))
#
#     return shadows


# class Shadow:
#     def __init__(self, edge_codes):
#         self.edge_codes = edge_codes
#         self.vertices = [edges for edges in edge_codes if len(edges) == 3]
#         self.crossings = [edges for edges in edge_codes if len(edges) == 4]
#         self.num_edges = sum(len(edges) for edges in edge_codes) // 2
#
#     def spatial_graph_diagram(self, signs=None, check=True):
#         num_cross = len(self.crossings)
#         if signs is None:
#             signs = num_cross * [0]
#         else:
#             assert len(signs) == num_cross
#
#         classes = [Edge(i) for i in range(self.num_edges)]
#         for v, edges in enumerate(self.vertices):
#             d = len(edges)
#             V = Vertex(d, 'V%d' % v)
#             classes.append(V)
#             for i, e in enumerate(edges):
#                 E = classes[e]
#                 e = 0 if E.adjacent[0] is None else 1
#                 V[i] = E[e]
#
#         for c, edges in enumerate(self.crossings):
#             C = Crossing('C%d' % c)
#             classes.append(C)
#             for i, e in enumerate(edges):
#                 E = classes[e]
#                 e = 0 if E.adjacent[0] is None else 1
#                 C[(i + signs[c]) % 4] = E[e]
#
#         return SpatialGraphDiagram(classes, check=check)


# def spatial_graph_diagrams_from_shadow(shadow, signs):
#    assert len(signs

# TODO Compile Plantri for Windows
# def spatial_graph_diagrams_fixed_crossings(G, crossings):
#     """
#     Let's start with the theta graph
#
#     >>> T = nx.MultiGraph(3*[(0, 1)])
#     >>> len(list(spatial_graph_diagrams_fixed_crossings(T, 3)))
#     2
#     """
#     assert all(d == 3 for v, d in G.degree)
#     assert all(a != b for a, b in G.edges())
#
#     raw_shadows = shadows_via_plantri_by_edge_codes(G.number_of_nodes(), crossings)
#
#     for raw_shadow in raw_shadows:
#         shadow = Shadow(raw_shadow)
#         diagram = shadow.spatial_graph_diagram(check=False)
#         U = diagram.underlying_graph()
#         if U is not None:
#             if nx.is_isomorphic(G, U):
#                 if not diagram.has_R6():
#                     num_cross = len(shadow.crossings)
#                     if num_cross == 0:
#                         yield diagram
#                     else:
#                         for signs in itertools.product((0, 1), repeat=num_cross - 1):
#                             signs = (0,) + signs
#                             D = shadow.spatial_graph_diagram(signs=signs, check=False)
#                             if not D.has_R2():
#                                 yield D








# def enumerate_yamada_classes(G, max_crossings):
#     examined = 0
#     polys = dict()
#     for crossings in range(0, max_crossings + 1):
#         for D in spatial_graph_diagrams_fixed_crossings(G, crossings):
#             p = D.yamada_polynomial()
#             p = normalize_poly(p)
#             if p not in polys:
#                 polys[p] = D
#             examined += 1
#     return polys, examined


# def to_poly(diagram):
#     p = diagram.yamada_polynomial()
#     p = normalize_poly(p)
#     return p, diagram


# def enumerate_yamada_classes_multicore(G, max_crossings, pool):
#     examined = 0
#     polys = dict()
#     timings = dict()
#     for crossings in range(0, max_crossings + 1):
#         start = time.time()
#         diagrams = spatial_graph_diagrams_fixed_crossings(G, crossings)
#         some_polys = pool.imap_unordered(to_poly, diagrams)
#         for p, D in some_polys:
#             if p not in polys:
#                 polys[p] = D
#             examined += 1
#         timings[crossings] = time.time() - start
#     return polys, timings, examined


# def pickle_yamada(data, filename):
#     with open(filename, 'wb') as file:
#         pickle.dump(data, file)


# def load_yamada(filename):
#     with open(filename, 'rb') as file:
#         return pickle.load(file)


"""Spatial Graphs

This module contains classes and functions for working with spatial graphs.
"""

import numpy as np
from numpy import sin, cos
import matplotlib.pyplot as plt
from itertools import combinations
from numba import njit
import math

from .spatial_graph_diagrams import (Vertex, Crossing, SpatialGraphDiagram)


class AbstractGraph:
    """
    TODO Figure out what happens if the graph has dangling edges
    """

    def __init__(self,
                 nodes: list[str],
                 edges: list[tuple[str, str]]):

        self.nodes = self._validate_nodes(nodes)
        self.edges = self._validate_edges(edges)

        self.adjacent_edge_pairs = self._get_adjacent_edge_pairs()
        self.nonadjacent_edge_pairs = [edge_pair for edge_pair in self.edge_pairs if edge_pair not in self.adjacent_edge_pairs]

    @staticmethod
    def _validate_nodes(nodes: list[str]) -> list[str]:
        """
        Validates the user's input and returns a list of nodes.

        Checks:
        1. The input is a list
        2. Each node is a string
        3. Each node string only contains alphanumeric characters
        4. All nodes are unique
        """

        if type(nodes) != list:
            raise TypeError('Nodes must be a list.')

        for node in nodes:
            if type(node) != str:
                raise TypeError('Nodes must be strings.')

        for node in nodes:
            if len(node) == 0:
                raise ValueError('Nodes must be a at least a single character.')

        if len(nodes) != len(set(nodes)):
            raise ValueError('All nodes must be unique.')

        # TODO Temporarily disabled for nodes like crossing "C1+" and "C1-".
        # for node in nodes:
        #     if not node.isalnum():
        #         raise ValueError('Nodes must be alphanumeric.')

        return nodes

    def _validate_edges(self, edges: list[tuple[str, str]]) -> list[tuple[str, str]]:
        """
        Validates the user's input and returns a list of edges.

        Checks:
        1. The input is a list
        2. Each edge is a tuple
        3. Each edge contains two valid nodes
        4. All edges are unique
        """

        if type(edges) != list:
            raise TypeError('Edges must be a list.')

        for edge in edges:
            if type(edge) != tuple:
                raise TypeError('Edge must be tuples.')

        for edge in edges:
            if len(edge) != 2:
                raise ValueError('Edges must contain two nodes.')

        for edge in edges:
            for node in edge:
                if node not in self.nodes:
                    raise ValueError('Edges must contain valid nodes.')

        if len(edges) != len(set(edges)):
            raise ValueError('All edges must be unique.')

        return edges

    def node_degree(self, node: str) -> int:
        """
        Returns the degree of a node.

        The node degree is the number of edges that are incident to the node.
        """
        return len([edge for edge in self.edges if node in edge])

    def get_adjacent_nodes(self, reference_node: str) -> list[str]:
        """
        Get the adjacent nodes to a given node.
        """

        adjacent_nodes = []

        for edge in self.edges:
            if reference_node == edge[0] or reference_node == edge[1]:
                adjacent_nodes += [node for node in edge if node != reference_node]

        return adjacent_nodes

    @property
    def edge_pairs(self):
        return list(combinations(self.edges, 2))

    def _get_adjacent_edge_pairs(self):

        adjacent_edge_pairs = []

        for edge_1, edge_2 in self.edge_pairs:

            a, b = edge_1
            c, d = edge_2

            assertion_1 = a in [c, d]
            assertion_2 = b in [c, d]

            if assertion_1 or assertion_2:
                adjacent_edge_pairs += [(edge_1, edge_2)]

        return adjacent_edge_pairs


class LinearAlgebra:
    """
    A class that contains static methods for necessary linear algebra calculations.
    """

    def __init__(self):
        self.rotated_node_positions = None
        self.rotation_generator_object = self.rotation_generator()
        self.rotation = self.random_rotation()

    @staticmethod
    def rotation_generator():
        """
        A generator to generate random rotations for projecting a spatial graph onto a 2D plane. Angles in radians.
        """

        # Set the initial generator state to zero.
        rotation = np.zeros(3)

        while True:
            yield rotation
            rotation = np.random.rand(3) * 2 * np.pi

    def random_rotation(self):
        """
        Query the rotation generator for a new rotation.
        """
        return next(self.rotation_generator_object)

    @staticmethod
    # @njit(cache=True)
    def rotate(positions: np.ndarray,
               rotation:  np.ndarray) -> np.ndarray:
        """
        Rotates a set of points about the first 3D point in the array.

        :param positions: A numpy array of 3D points.
        :param rotation: A numpy array of 3 Euler angles in radians.

        :return: new_positions:
        """

        # Shift the object to origin
        reference_position = positions[0]
        origin_positions = positions - reference_position

        alpha, beta, gamma = rotation

        # Rotation matrix Euler angle convention r = r_z(gamma) @ r_y(beta) @ r_x(alpha)

        r_x = np.array([[1., 0., 0.],
                        [0., cos(alpha), -sin(alpha)],
                        [0., sin(alpha), cos(alpha)]])

        r_y = np.array([[cos(beta), 0., sin(beta)],
                        [0., 1., 0.],
                        [-sin(beta), 0., cos(beta)]])

        r_z = np.array([[cos(gamma), -sin(gamma), 0.],
                        [sin(gamma), cos(gamma), 0.],
                        [0., 0., 1.]])

        # TODO Replace with explicitly formulated matrix
        r = r_z @ r_y @ r_x

        # Transpose positions from [[x1,y1,z1],[x2... ] to [[x1,x2,x3],[y1,... ]
        rotated_origin_positions = (r @ origin_positions.T).T

        # Shift back from origin
        new_positions = rotated_origin_positions + reference_position

        rotated_node_positions = new_positions

        return rotated_node_positions


    @staticmethod
    @njit(cache=True)
    def get_line_segment_intersection(a: np.ndarray,
                                      b: np.ndarray,
                                      c: np.ndarray,
                                      d: np.ndarray) -> tuple[float, np.ndarray]:
        """
        Returns the minimum Euclidean distance between two line segments and its position.

        This function also works for calculating the distance between a line segment and a point and a point and point.

        Based on the algorithm described in:

        Vladimir J. Lumelsky,
        "On Fast Computation of Distance Between Line Segments",
        Information Processing Letters 21 (1985) 55-61
        https://doi.org/10.1016/0020-0190(85)90032-8

        Values 0 <= t <= 1 correspond to points being inside segment AB whereas values < 0  correspond to being 'left' of AB
        and values > 1 correspond to being 'right' of AB.

        Values 0 <= u <= 1 correspond to points being inside segment CD whereas values < 0  correspond to being 'left' of CD
        and values > 1 correspond to being 'right' of CD.

        Step 1: Check for special cases; compute D1, D2, and the denominator in (11)
            (a) If one of the two segments degenerates into a point, assume that this segment corresponds to the parameter
            u, take u=0, and go to Step 4.
            (b) If both segments degenerate into points, take t=u=0, and go to Step 5.
            (c) If neither of two segments degenerates into a point and the denominator in (11) is zero, take t=0 and go to
            Step 3.
            (d) If none of (a), (b), (c) takes place, go to Step 2.
        Step 2: Using (11) compute t. If t is not in the range [0,1], modify t using (12).
        Step 3: Using (10) compute u. If u is not in the range [0,1], modify u using (12); otherwise, go to Step 5.
        Step 4: Using (10) compute t. If t is not in the range [0,1], modify t using (12).
        Step 5: With current values of t and u, compute the actual MinD using (7).

        :param a: (1,3) numpy array
        :param b: (1,3) numpy array
        :param c: (1,3) numpy array
        :param d: (1,3) numpy array

        :return: Minimum distance between line segments, float
        :return: Position of minimum distance, (1,3) numpy array
        """

        def clamp_bound(num):
            """
            If the number is outside the range [0,1] then clamp it to the nearest boundary.
            """
            if num < 0.:
                return 0.
            elif num > 1.:
                return 1.
            else:
                return num

        d1  = b - a
        d2  = d - c
        d12 = c - a

        D1  = np.dot(d1, d1.T)
        D2  = np.dot(d2, d2.T)
        S1  = np.dot(d1, d12.T)
        S2  = np.dot(d2, d12.T)
        R   = np.dot(d1, d2.T)
        den = D1 * D2 - R**2

        # Check if one or both line segments are points
        if D1 == 0. or D2 == 0.:

            # Both AB and CD are points
            if D1 == 0. and D2 == 0.:
                t = 0.
                u = 0.

            # AB is a line segment and CD is a point
            elif D1 != 0.:
                u = 0.
                t = S1 / D1
                t = clamp_bound(t)

            # AB is a point and CD is a line segment
            elif D2 != 0.:
                t = 0.
                u = -S2 / D2
                u = clamp_bound(u)

        # Check if line segments are parallel
        elif den == 0.:
            t = 0.
            u = -S2 / D2
            uf = clamp_bound(u)

            if uf != u:
                t = (uf * R + S1) / D1
                t = clamp_bound(t)
                u = uf

        # General case for calculating the minimum distance between two line segments
        else:

            t = (S1 * D2 - S2 * R) / den

            t = clamp_bound(t)

            u = (t * R - S2) / D2
            uf = clamp_bound(u)

            if uf != u:
                t = (uf * R + S1) / D1
                t = clamp_bound(t)

                u = uf

        min_dist = np.linalg.norm(d1 * t - d2 * u - d12)

        # min_dist_position = np.array([a + d1 * t]).reshape(-1)
        min_dist_position = a + d1 * t

        return min_dist, min_dist_position


    @staticmethod
    # @njit(cache=True)
    def calculate_intermediate_y_position(a:     np.ndarray,
                                          b:     np.ndarray,
                                          x_int: float,
                                          z_int: float) -> float:
        """
        Calculates the intermediate y position given two points and the intermediate x and z position.
        """

        x1, y1, z1 = a
        x2, y2, z2 = b

        delta_x = x2 - x1
        delta_y = y2 - y1
        # delta_z = z2 - z1

        ratio_x = (x_int - x1) / delta_x
        # ratio_z = (z_int - z1) / delta_z
        # assert np.isclose(ratio_x, ratio_z, rtol=1e-3)

        y_int = y1 + ratio_x * delta_y

        return y_int

    @property
    def projected_node_positions(self):
        return self.rotated_node_positions[:, [0, 2]]


    @staticmethod
    # @njit(cache=True)
    def calculate_counter_clockwise_angle(vector_a: np.ndarray,
                                          vector_b: np.ndarray) -> float:
        """
        Returns the angle in degrees between vectors 'A' and 'B'.

        :param vector_a: A numpy array of shape (2,) representing containing positions x_a and y_a.
        :param vector_b: A numpy array of shape (2,) representing containing positions x_b and y_b.
        :return: The angle in degrees between vectors A and B, where 0 <= angle < 360.
        """

        def length(v):
            return np.sqrt(v[0] ** 2 + v[1] ** 2)

        def dot_product(v, w):
            return v[0] * w[0] + v[1] * w[1]

        def determinant(v, w):
            return v[0] * w[1] - v[1] * w[0]

        def inner_angle(v, w):
            cosx = dot_product(v, w) / (length(v) * length(w))
            rad = np.arccos(cosx)     # in radians
            return rad * 180 / np.pi  # returns degrees

        inner = inner_angle(vector_a, vector_b)
        det = determinant(vector_a, vector_b)

        if det > 0:  # this is a property of the det. If the det < 0 then B is clockwise of A
            return inner
        else:  # if the det > 0 then A is immediately clockwise of B
            return 360 - inner


class SpatialGraph(AbstractGraph, LinearAlgebra):
    """
    A class to represent a spatial graph.

    Notes:
        1. Component ports must be individually represented as nodes. In other words, a control valve servicing an
        actuator must employ separate nodes for the supply and return ports, along with their associated edges.
    """

    def __init__(self,
                 nodes: list[str],
                 node_positions: np.ndarray,
                 edges: list[tuple[str, str]]):

        # Initialize AbstractGraph Class
        AbstractGraph.__init__(self, nodes, edges)

        # Initialize Node positions
        self.node_positions = self._validate_node_positions(node_positions)

        # Initialize attributes necessary for geometric calculations
        LinearAlgebra.__init__(self)

        # Initialize a first rotation
        self.rotated_node_positions = self.rotate(self.node_positions, self.rotation)

        # Initialize attributes that are calculated later
        self.crossings = None
        self.crossing_positions = None
        self.crossing_edge_pairs = None

        # Project the spatial graph onto a random the xz-plane
        self.project()

    def _validate_node_positions(self,
                                 node_positions: np.ndarray) -> np.ndarray:
        """
        Validates the user's input and returns an array of node positions.

        Checks:
        1. The input is a np.ndarray
        2. The array size matches the number of nodes
        3. Each row contains 3 real numbers
        4. Each row is unique
        """

        if type(node_positions) != np.ndarray:

            # TODO Check for if elements in list are not numbers, etc.
            if type(node_positions) == list:
                node_positions = np.array(node_positions)

            else:
                raise TypeError('Node positions must be a numpy array.')

        if node_positions.shape[0] != len(self.nodes):
            raise ValueError('Node positions must contain a position for each node.')

        if node_positions.shape[1] != 3:
            raise ValueError('Node positions must contain 3D coordinates.')

        valid_types = [float, int, np.float32, np.int32, np.float64, np.int64]
        for row in node_positions:
            for element in row:
                if type(element) not in valid_types:
                    raise TypeError('Node positions must contain real numbers.')

        if node_positions.shape[0] != np.unique(node_positions, axis=0).shape[0]:
            raise ValueError('All nodes must have a position.')

        return node_positions

    def get_vertices_and_crossings_of_edge(self, reference_edge: tuple[str, str]):
        """
        Returns the vertices and crossings of an edge, ordered from left to right.
        """

        # Get edge nodes and positions
        edge_nodes = [node for node in reference_edge]
        edge_node_positions = [self.projected_node_positions[self.nodes.index(node)] for node in edge_nodes]

        # Get crossing and positions (if applicable)
        edge_crossings = []
        edge_crossing_positions = []

        for edge_pair, position, crossing in zip(self.crossing_edge_pairs, self.crossing_positions, self.crossings):
            if reference_edge in edge_pair:
                edge_crossings.append(crossing)
                edge_crossing_positions.append(position)

        if len(edge_crossings) > 0:
            # Merge the vertices and crossings
            adjacent_nodes = edge_nodes + edge_crossings
            adjacent_positions = np.vstack((edge_node_positions, edge_crossing_positions))

        else:
            adjacent_nodes = edge_nodes
            adjacent_positions = edge_node_positions

        # Order vertices and crossings from left to right by x position (i.e., ascending position index 0)
        ordered_nodes = [node for _, node in
                         sorted(zip(adjacent_positions, adjacent_nodes), key=lambda pair: pair[0][0])]

        return ordered_nodes

    def get_adjacent_nodes_or_crossings(self, reference_node):
        """
        Get the adjacent nodes to a given node.
        """

        adjacent_nodes = []

        for edge in self.get_sub_edges():
            if reference_node == edge[0] or reference_node == edge[1]:
                adjacent_nodes += [node for node in edge if node != reference_node]

        return adjacent_nodes

    def get_adjacent_nodes_projected_positions(self, reference_node: str):
        """
        Get the adjacent nodes to a given node.
        """
        adjacent_nodes = self.get_adjacent_nodes(reference_node)

        node_indices = [self.nodes.index(node) for node in adjacent_nodes]

        return self.projected_node_positions[node_indices]

    def cyclic_order_vertex(self,
                            reference_node:     str,
                            node_ordering_dict: dict = None) -> dict:
        """
        Get the cyclical order of adjacent vertexes and crossings for a given vertex.

        The spatial-topological calculations presume that nodes are ordered in a CCW rotation. While it does not matter
        which node is used as index zero, it is important that the node order is consistent.

        :param reference_node: The node to use as the reference node.
        :param node_ordering_dict: A dictionary of the node order.
        :return: A dictionary of the node order.
        """

        # If no node ordering dictionary is provided, use the default node ordering dictionary
        if node_ordering_dict is None:
            node_ordering_dict = {}

        # Get the projected node positions
        reference_node_position = self.projected_node_positions[self.nodes.index(reference_node)]

        # Initialize lists to store the adjacent node and edge information
        # Crossings are not relevant for this calculation since they exist along edges
        adjacent_nodes = self.get_adjacent_nodes_or_crossings(reference_node)
        adjacent_node_positions = self.get_adjacent_nodes_projected_positions(reference_node)

        # Shift nodes to the origin
        shifted_adjacent_node_positions = adjacent_node_positions - reference_node_position

        # Horizontal line (reference for angles)
        reference_vector = np.array([1, 0])

        rotations = []
        for shifted_adjacent_node_position in shifted_adjacent_node_positions:
            rotations.append(self.calculate_counter_clockwise_angle(reference_vector, shifted_adjacent_node_position))

        ordered_nodes = [node for _, node in sorted(zip(rotations, adjacent_nodes))]

        ccw_node_ordering = {}
        for node in ordered_nodes:
            ccw_node_ordering[node] = ordered_nodes.index(node)

        node_ordering_dict[reference_node] = ccw_node_ordering

        return node_ordering_dict

    def cyclic_order_vertices(self):
        node_ordering_dict = {}
        for node in self.nodes:
            node_ordering_dict = self.cyclic_order_vertex(node, node_ordering_dict)
        return node_ordering_dict

    def get_sub_edges(self):
        """
        Sub-edges constitute:
        1. Edges that are not incident of a crossing (i.e., the edge is the sub-edge)
        2. When an edge is incident to a crossing(s), the edge is divided into one or more sub-edges depending on the
        number of crossings that the edge is incident to.
        """

        # Get the nodes and crossings of each edge, ordered from left to right
        edge_nodes_and_crossings = {}
        for edge in self.edges:
            edge_nodes_and_crossings[edge] = self.get_vertices_and_crossings_of_edge(edge)

        # Subdivide each edge if necessary.
        # If there are 2 nodes then there is one sub-edge, 3 nodes means 2 sub-edges, etc.

        sub_edges = []

        for edge, nodes_and_crossings in edge_nodes_and_crossings.items():

            if len(nodes_and_crossings) == 2:
                sub_edges.append(edge)

            else:
                sub_edges += [(nodes_and_crossings[i], nodes_and_crossings[i + 1]) for i in
                              range(len(nodes_and_crossings) - 1)]

        return sub_edges


    def get_node_or_crossing_projected_position(self, reference_node: str) -> np.ndarray:

        if reference_node in self.nodes:
            return self.projected_node_positions[self.nodes.index(reference_node)]

        elif reference_node in self.crossings:
            return self.crossing_positions[self.crossings.index(reference_node)]

        else:
            raise ValueError("Node or crossing not found.")


    def cyclic_order_crossing(self,
                              crossing:           str,
                              node_ordering_dict: dict = None):
        """
        Get the order of the overlapping nodes in the two edges. First is under, second is over.

        Use rotated node positions since rotation and project may change the overlap order.

        :param crossing: The crossing
        :param node_ordering_dict: A dictionary of node ordering for each node
        :return: overlap_order: The order of the overlapping nodes in edge 1 and edge 2
        """

        # If no node ordering dictionary is provided, use the default node ordering dictionary
        if node_ordering_dict is None:
            node_ordering_dict = {}

        # Get the edges that are connected to the crossing
        # Assumption: get_edges_of_crossing returns the edges in the order of the nodes
        edge_1, edge_2 = self.crossing_edge_pairs[self.crossings.index(crossing)]
        edge_1_nodes_and_crossings = self.get_vertices_and_crossings_of_edge(edge_1)
        edge_2_nodes_and_crossings = self.get_vertices_and_crossings_of_edge(edge_2)

        # Get the nodes that are adjacent to the crossing (whether a vertex or another crossing)
        # Assumption: The crossing is not the first or last node of the edge
        crossing_index_edge_1 = edge_1_nodes_and_crossings.index(crossing)
        crossing_index_edge_2 = edge_2_nodes_and_crossings.index(crossing)

        edge_1_left_node  = edge_1_nodes_and_crossings[crossing_index_edge_1 - 1]
        edge_1_right_node = edge_1_nodes_and_crossings[crossing_index_edge_1 + 1]
        edge_2_left_node  = edge_2_nodes_and_crossings[crossing_index_edge_2 - 1]
        edge_2_right_node = edge_2_nodes_and_crossings[crossing_index_edge_2 + 1]


        # Get the vertices that the beginning and end of each edge
        # While the crossing might be adjoined be other crossings, crossings only occur in 2D.
        # The 3D positions of the vertices are required to determine which edge is in front of the other.

        # Determine which vertex is the left and right vertex of each edge.
        # Assumption: Edges are ordered nodes left to right

        edge_1_left_vertex  = edge_1_nodes_and_crossings[0]
        edge_1_right_vertex = edge_1_nodes_and_crossings[-1]
        edge_2_left_vertex  = edge_2_nodes_and_crossings[0]
        edge_2_right_vertex = edge_2_nodes_and_crossings[-1]

        edge_1_left_vertex_position  = self.rotated_node_positions[self.nodes.index(edge_1_left_vertex)]
        edge_1_right_vertex_position = self.rotated_node_positions[self.nodes.index(edge_1_right_vertex)]
        edge_2_left_vertex_position  = self.rotated_node_positions[self.nodes.index(edge_2_left_vertex)]
        edge_2_right_vertex_position = self.rotated_node_positions[self.nodes.index(edge_2_right_vertex)]

        crossing_position = self.crossing_positions[self.crossings.index(crossing)]

        y_crossing_edge_1 = self.calculate_intermediate_y_position(edge_1_left_vertex_position,
                                                                   edge_1_right_vertex_position,
                                                                   crossing_position[0],
                                                                   crossing_position[1])

        y_crossing_edge_2 = self.calculate_intermediate_y_position(edge_2_left_vertex_position,
                                                                   edge_2_right_vertex_position,
                                                                   crossing_position[0],
                                                                   crossing_position[1])

        # Get the projected node positions
        reference_node_position = self.get_node_or_crossing_projected_position(crossing)

        # Initialize lists to store the adjacent node and edge information
        # Crossings are not relevant for this calculation since they exist along edges

        edge_1_left_vertex_position_projected  = self.get_node_or_crossing_projected_position(edge_1_left_vertex)
        edge_1_right_vertex_position_projected = self.get_node_or_crossing_projected_position(edge_1_right_vertex)
        edge_2_left_vertex_position_projected  = self.get_node_or_crossing_projected_position(edge_2_left_vertex)
        edge_2_right_vertex_position_projected = self.get_node_or_crossing_projected_position(edge_2_right_vertex)

        adjacent_vertex_positions_projected = [edge_1_left_vertex_position_projected, edge_1_right_vertex_position_projected, edge_2_left_vertex_position_projected, edge_2_right_vertex_position_projected]

        adjacent_nodes    = [edge_1_left_node, edge_1_right_node, edge_2_left_node, edge_2_right_node]
        adjacent_vertices = [edge_1_left_vertex, edge_1_right_vertex, edge_2_left_vertex, edge_2_right_vertex]

        # Shift nodes to the origin
        shifted_adjacent_vertex_positions = adjacent_vertex_positions_projected - reference_node_position

        # Horizontal line (reference for angles)
        reference_vector = np.array([1, 0])

        rotations = []

        # Convention: +y goes into the screen and -y goes out of the screen
        # Therefore the edge that is in front of the other is the edge that has the lower y value
        if y_crossing_edge_2 < y_crossing_edge_1:
            under_edge = edge_1
            over_edge  = edge_2
        elif y_crossing_edge_2 > y_crossing_edge_1:
            under_edge = edge_2
            over_edge  = edge_1
        else:
            raise Exception('These crossings should not intersect in 3D')

        # Calculate the angle of each node
        for shifted_adjacent_node_position in shifted_adjacent_vertex_positions:
            rotations.append(self.calculate_counter_clockwise_angle(reference_vector, shifted_adjacent_node_position))

        # Sort the nodes by their angle
        ordered_nodes = [node for _, node in sorted(zip(rotations, adjacent_nodes))]

        # Sort the vertices by their angle (used for reference)
        ordered_vertices = [node for _, node in sorted(zip(rotations, adjacent_vertices))]

        # Check if the first node after the reference vector is on the under edge or the over edge
        first_vertex = ordered_vertices[0]
        if first_vertex in under_edge:
            crossing_indices = [0, 1, 2, 3]
        elif first_vertex in over_edge:
            crossing_indices = [1, 2, 3, 0]
        else:
            raise Exception("First node is not on either edge")

        # Assign the cyclic ordering
        ccw_crossing_ordering = {}
        for node, index in zip(ordered_nodes, crossing_indices):
            ccw_crossing_ordering[node] = index

        node_ordering_dict[crossing] = ccw_crossing_ordering

        return node_ordering_dict

    def cyclic_order_crossings(self):
        """
        Get the node ordering of the graph with crossings.

        TODO Replace crossing indices with crossings
        """
        # Create a dictionary that contains the cyclical ordering of every crossing
        crossing_ordering_dict = {}
        for crossing in self.crossings:
            crossing_ordering_dict = self.cyclic_order_crossing(crossing, crossing_ordering_dict)
        return crossing_ordering_dict

    def get_crossing_edge_order(self,
                                edge_1,
                                edge_2,
                                crossing_position):
        """
        Get the order of the overlapping nodes in the two edges. First is under, second is over.

        Use rotated node positions since rotation and project may change the overlap order.

        :param edge_1: Edge 1
        :param edge_2: Edge 2
        :param crossing_position: The point of intersection between the projections of edge 1 and edge 2

        :return: overlap_order: The order of the overlapping nodes in edge 1 and edge 2
        """

        overlap_order = []

        node_1, node_2 = edge_1
        node_3, node_4 = edge_2

        node_1_index = self.nodes.index(node_1)
        node_2_index = self.nodes.index(node_2)
        node_3_index = self.nodes.index(node_3)
        node_4_index = self.nodes.index(node_4)

        node_1_position = self.rotated_node_positions[node_1_index]
        node_2_position = self.rotated_node_positions[node_2_index]
        node_3_position = self.rotated_node_positions[node_3_index]
        node_4_position = self.rotated_node_positions[node_4_index]

        x_crossing, z_crossing = crossing_position

        y_crossing_1 = self.calculate_intermediate_y_position(node_1_position, node_2_position, x_crossing, z_crossing)
        y_crossing_2 = self.calculate_intermediate_y_position(node_3_position, node_4_position, x_crossing, z_crossing)

        # todo Only check if crossing is in bounds!
        if y_crossing_1 == y_crossing_2:
            raise RuntimeError('The edges are planar and therefore the interconnects they represent physically '
                               'intersect. This is not a valid spatial graph. Edges: {}, {}.'.format(edge_1, edge_2))

        elif y_crossing_1 > y_crossing_2:
            overlap_order.append(edge_2)
            overlap_order.append(edge_1)

        elif y_crossing_1 < y_crossing_2:
            overlap_order.append(edge_1)
            overlap_order.append(edge_2)

        else:
            raise NotImplementedError('There should be no else case.')

        # Convert list to tuple for hashing later on...
        overlap_order = tuple(overlap_order)

        return overlap_order


    # @njit
    def get_crossings(self):

        crossing_num = 0
        crossings    = []
        crossing_edge_pairs = []
        crossing_positions = []

        # TODO Modify nonadjacent edge pairs that are within some axis aligned bounding box




        for line_1, line_2 in self.nonadjacent_edge_pairs:

            a = self.projected_node_positions[self.nodes.index(line_1[0])]
            b = self.projected_node_positions[self.nodes.index(line_1[1])]
            c = self.projected_node_positions[self.nodes.index(line_2[0])]
            d = self.projected_node_positions[self.nodes.index(line_2[1])]

            min_dist, crossing_position = self.get_line_segment_intersection(a, b, c, d)

            if min_dist > 0.0001:
                crossing_position = None

            if crossing_position is np.inf:
                raise ValueError('The edges are overlapping. This is not a valid spatial graph.')

            elif crossing_position is None:
                pass

            elif type(crossing_position) is np.ndarray and min_dist < 0.0001:
                crossings.append('crossing_' + str(crossing_num))
                crossing_num += 1
                crossing_positions.append(crossing_position)
                crossing_edge_pair = self.get_crossing_edge_order(line_1, line_2, crossing_position)
                crossing_edge_pairs.append(crossing_edge_pair)

        return crossings, crossing_positions, crossing_edge_pairs

    def project(self, max_iter=50, predefined_rotation=None):
        """
        Project the spatial graph onto a random 2D plane.

        The projection is done by applying a random 3D rotation to the graph and then projecting it onto a 2D plane.
        For simplicity, we use the transformed XZ plane. When the rotation is complete, the program checks that the
        projected graph for several things.

        First, it checks to make sure that no combinations of edges and/or nodes are overlapping.

        Second, it checks to make sure that no edges are perfectly vertical or perfectly horizontal. While neither of
        these cases technically incorrect, it's easier to implement looping through rotations rather than add edge cases
        in for these cases.

        """

        if predefined_rotation is not None:
            self.rotation = predefined_rotation
            self.rotated_node_positions = self.rotate(self.node_positions, self.rotation)
            self.crossings, self.crossing_positions, self.crossing_edge_pairs = self.get_crossings()
            return


        rotations = []
        num_crossings = []

        for _ in range(max_iter):

            try:

                # First, check that no edges are perfectly vertical or perfectly horizontal.
                # While neither of these cases technically incorrect, it's easier to implement looping through rotations
                # rather than add edge cases for each 2D and 3D line equation.

                for edge in self.edges:
                    x1, z1 = self.projected_node_positions[self.nodes.index(edge[0])]
                    x2, z2 = self.projected_node_positions[self.nodes.index(edge[1])]

                    if x1 == x2 or z1 == z2:
                        raise ValueError('An edge is vertical or horizontal. This is not a valid spatial graph.')

                # Second, check adjacent edge pairs for validity.
                # Since adjacent segments are straight lines, they should only intersect at a single endpoint.
                # The only other possibility is for them to infinitely overlap, which is not a valid spatial graph.

                for line_1, line_2 in self.adjacent_edge_pairs:
                    a = self.projected_node_positions[self.nodes.index(line_1[0])]
                    b = self.projected_node_positions[self.nodes.index(line_1[1])]
                    c = self.projected_node_positions[self.nodes.index(line_2[0])]
                    d = self.projected_node_positions[self.nodes.index(line_2[1])]
                    min_dist, crossing_position = self.get_line_segment_intersection(a, b, c, d)

                    if min_dist > 0.0001:
                        crossing_position = None

                    if crossing_position is not None and crossing_position is not np.inf:
                        assertion_1 = all((crossing_position - a) < 0.0001)
                        assertion_2 = all((crossing_position - b) < 0.0001)
                        assertion_3 = all((crossing_position - c) < 0.0001)
                        assertion_4 = all((crossing_position - d) < 0.0001)
                        if not any([assertion_1, assertion_2, assertion_3, assertion_4]):
                            raise ValueError('Adjacent edges must intersect at the endpoints.')

                    elif crossing_position is np.inf:
                        raise ValueError('The edges are overlapping. This is not a valid spatial graph.')


                # Third, Check nonadjacent edge pairs for validity.
                # Since nonadjacent segments are straight lines, they should only intersect at zero or one points.
                # Since adjacent segments should only overlap at endpoints, nonadjacent segments should only overlap
                # between endpoints.
                # The only other possibility is for them to infinitely overlap, which is not a valid spatial graph.

                self.crossings, self.crossing_positions, self.crossing_edge_pairs = self.get_crossings()

                rotations.append(self.rotation)
                num_crossings.append(len(self.crossings))

            except ValueError:
                self.rotation = self.random_rotation()
                self.rotated_node_positions = self.rotate(self.node_positions, self.rotation)

            # If there is no error in the projection rotate anyways
            self.rotation = self.random_rotation()
            self.rotated_node_positions = self.rotate(self.node_positions, self.rotation)



        if len(rotations) == 0:
            raise Exception('Could not find a valid rotation after {} iterations'.format(max_iter))
        else:
            min_crossings = min(num_crossings)
            min_crossings_index = num_crossings.index(min_crossings)
            self.rotation = rotations[min_crossings_index]
            self.rotated_node_positions = self.rotate(self.node_positions, self.rotation)
            self.crossings, self.crossing_positions, self.crossing_edge_pairs = self.get_crossings()


    def create_vertices(self):
        """
        Create the vertices of the spatial graph.
        """

        vertices = [Vertex(self.node_degree(node), 'node_' + node) for node in self.nodes]

        return vertices

    def create_crossings(self):
        """
        Create the crossings of the spatial graph.
        """

        if self.crossing_positions is not None:
            crossings = [Crossing('crossing_object_' + str(i)) for i in range(len(self.crossing_positions))]
        else:
            crossings = []

        return crossings

    def create_spatial_graph_diagram(self):
        """
        Create a diagram of the spatial graph.
        """

        nodes_and_crossings = self.nodes + self.crossings

        # Create the vertex and crossing objects
        vertices = self.create_vertices()
        crossings = self.create_crossings()
        vertices_and_crossings = vertices + crossings

        # Create a dictionary that contains the cyclical ordering of every node and crossing
        node_ordering_dict = self.cyclic_order_vertices()
        crossing_ordering_dict = self.cyclic_order_crossings()

        cyclic_ordering_dict = {**node_ordering_dict, **crossing_ordering_dict}

        for sub_edge in self.get_sub_edges():
            node_a, node_b = sub_edge

            node_a_index = nodes_and_crossings.index(node_a)
            node_b_index = nodes_and_crossings.index(node_b)

            vertex_a = vertices_and_crossings[node_a_index]
            vertex_b = vertices_and_crossings[node_b_index]

            if not vertex_a.already_assigned(vertex_b) and not vertex_b.already_assigned(vertex_a):

                vertex_b_index_for_vertex_a = cyclic_ordering_dict[node_a][node_b]
                vertex_a_index_for_vertex_b = cyclic_ordering_dict[node_b][node_a]

                vertex_a[vertex_b_index_for_vertex_a] = vertex_b[vertex_a_index_for_vertex_b]

            else:
                raise ValueError('The vertices are already assigned.')

        inputs = vertices + crossings

        sgd = SpatialGraphDiagram(inputs)

        return sgd

    def plot(self):
        """
        Plot the spatial graph and spatial graph diagram.
        """

        fig = plt.figure()
        # ax2 = plt.subplot()

        ax1 = plt.subplot(221, projection='3d')
        ax2 = plt.subplot(222)
        # ax3 = plt.subplot(223, projection='3d')
        #
        # # Axis 1
        #
        # # ax1.set_xlim(-1.5, 2)
        # # ax1.set_ylim(-1.5, 2)
        # # ax1.set_zlim(-1.5, 2)
        #
        # ax1.title.set_text('Spatial Graph')
        # ax1.xaxis.label.set_text('x')
        # ax1.yaxis.label.set_text('y')
        # ax1.zaxis.label.set_text('z')
        #
        # ax1.set_yticklabels([])
        # ax1.set_xticklabels([])
        #
        # # Axis 2
        #
        # ax2.title.set_text('Spatial Graph Diagram 1 \n (XZ Plane Projection)')
        # ax2.xaxis.label.set_text('x')
        # ax2.yaxis.label.set_text('z')
        #
        # ax2.set_yticklabels([])
        # ax2.set_xticklabels([])
        #
        # # ax2.set_xlim(-0.5, 1.5)
        # # ax2.set_ylim(-0.5, 1.5)
        #
        # # Axis 3
        #
        # # ax3.set_xlim(-1.25, 2)
        # # ax3.set_ylim(-1.25, 2)
        # # ax3.set_zlim(-1.25, 2)
        #
        # ax3.title.set_text('Spatial Graph Rotated')
        # ax3.xaxis.label.set_text('x')
        # ax3.yaxis.label.set_text('y')
        # ax3.zaxis.label.set_text('z')
        #
        # # Figure layout
        # # plt.tight_layout(pad=2, w_pad=7, h_pad=0)
        #
        # Plot 3D
        for edge in self.edges:
            point_1 = self.node_positions[self.nodes.index(edge[0])]
            point_2 = self.node_positions[self.nodes.index(edge[1])]
            ax1.plot3D([point_1[0], point_2[0]], [point_1[1], point_2[1]], [point_1[2], point_2[2]])
        #
        # # Shrink current axis by 40%
        # box = ax1.get_position()
        # ax1.set_position([box.x0, box.y0, box.width * 0.75, box.height])
        #
        # # Put a legend to the right of the current axis
        # # ax1.legend(self.edges, loc='center left', bbox_to_anchor=(1.5, 0.5))
        #
        # # Plot 3D
        # for edge in self.edges:
        #     point_1 = self.rotated_node_positions[self.nodes.index(edge[0])]
        #     point_2 = self.rotated_node_positions[self.nodes.index(edge[1])]
        #     ax3.plot3D([point_1[0], point_2[0]], [point_1[1], point_2[1]], [point_1[2], point_2[2]])
        # # ax3.legend(self.edges)
        #
        # # Shrink current axis by 40%
        # box = ax3.get_position()
        # ax3.set_position([box.x0, box.y0, box.width * 0.75, box.height])

        # Put a legend to the right of the current axis
        # ax3.legend(self.edges, loc='center left', bbox_to_anchor=(1.5, 0.5))
        # ax3.legend(self.edges)

        # Plot 2D
        for edge in self.edges:
            point_1 = self.projected_node_positions[self.nodes.index(edge[0])]
            point_2 = self.projected_node_positions[self.nodes.index(edge[1])]
            ax2.plot([point_1[0], point_2[0]], [point_1[1], point_2[1]])
        # ax2.legend(self.edges)

        # Shrink current axis by 40%
        box = ax2.get_position()
        # ax2.set_position([box.x0, box.y0, box.width * 0.6, box.height])

        # Put a legend to the right of the current axis
        # ax2.legend(self.edges, loc='center left', bbox_to_anchor=(1, -0.25))
        # ax2.legend(self.edges, loc='center left', bbox_to_anchor=(0.75, 0.5), fontsize=9)
        # ax2.legend(self.edges)


        # Plot crossing positions
        if self.crossing_positions is not None:
            for crossing_position in self.crossing_positions:
                ax2.scatter(crossing_position[0], crossing_position[1],
                            marker='o', s=500, facecolors='none', edgecolors='r', linewidths=2)

        plt.show()



# TODO Implement

# import networkx as nx
# from networkx.algorithms.planar_drawing import triangulate_embedding
# import numpy as np
# from collections import Counter
# from sage.all import sphere, line3d, polygon3d, colormaps, bezier3d

# def choose_outer_face(planar_graph):
#     G = planar_graph
#     half_edges = set(G.edges)
#     faces = []
#     while len(half_edges):
#         a, b = half_edges.pop()
#         face_edges = set()
#         face_verts = G.traverse_face(a, b, face_edges)
#         half_edges.difference_update(face_edges)        
#         faces.append(face_verts)

#     print(len(faces))
#     def score_face(face):
#         counts = Counter(G.degree[v] for v in face)
#         return counts[8]

#     # return [(score_face(face), face) for face in faces]
#     return min(faces, key=score_face)


# def tutte_system(planar_graph):
#     G, outer = triangulate_embedding(planar_graph, fully_triangulate=False)
#     #G = planar_graph
#     node_to_index = {node:index for index, node in enumerate(G.nodes())}
#     #outer = choose_outer_face(planar_graph)
#     n = G.number_of_nodes()
#     assert n == planar_graph.number_of_nodes()
#     inner_indices = np.array([node not in outer for node in G.nodes()])
#     L = nx.laplacian_matrix(G).toarray()
#     L = L[inner_indices, :]
#     M = np.zeros((len(outer), n))
#     for i, node in enumerate(outer):
#         M[i, node_to_index[node]] = 1
#     A = np.vstack([L, M])
#     Z = np.zeros(A.shape)
#     B = np.block([[A, Z], [Z, A]])
#     # set locations of boundary nodes
#     t = np.linspace(0, 2*np.pi, len(outer), endpoint=False)
#     x, y = 100*np.cos(t), 100*np.sin(t)
#     zeros = np.zeros(L.shape[0])
#     b = np.hstack([zeros, x, zeros, y])
#     return G, B, b

# def tutte_embedding_positions(planar_graph):
#     n = planar_graph.number_of_nodes()
#     G, A, b = tutte_system(planar_graph)
#     pos = np.linalg.solve(A, b)
#     pos = np.rint(pos)
#     x, y = pos[:n], pos[n:]
#     ans = dict()
#     for i, node in enumerate(G.nodes()):
#         ans[node] = (int(x[i]), int(y[i]))
#     return ans


# def position_spatial_graph_in_3D(G, z_height=20):
#     def norm_label(X):
#         L = X.label
#         return repr(L) if not isinstance(L, str) else L
        
#     P = G.underlying_planar_embedding()
#     planar_pos = tutte_embedding_positions(P)
#     system_node_pos = dict()
#     for V in G.vertices:
#         L = norm_label(V)
#         x, y = planar_pos[V.label]
#         system_node_pos[L] = (x, y, 0)

#     other_node_pos = dict()
#     for C in G.crossings:
#         L = norm_label(C)
#         x, y = planar_pos[C.label]
#         other_node_pos[L + '+'] = (x, y, z_height)
#         other_node_pos[L + '-'] = (x, y, -z_height)

#     nodes_so_far = system_node_pos.copy()
#     nodes_so_far.update(other_node_pos)

#     def end_label(edge, i):
#         i = i % 2
#         X, x = edge.adjacent[i]
#         L = norm_label(X)
#         if X in G.crossings:
#             L = L + '-' if x % 2 == 0 else L + '+'
#         return L

#     for E in G.edges:
#         L = norm_label(E)
#         x, y = planar_pos[E.label]
#         A = end_label(E, 0)
#         B = end_label(E, 1)
#         z = (nodes_so_far[A][2] + nodes_so_far[B][2]) // 2        
#         other_node_pos[L] = (x, y, z)
        
#     segments = list()
#     vertex_inputs = set()
#     for V in G.vertices:
#         vertex_inputs.update((V, i) for i in range(V.degree))
#     while len(vertex_inputs):
#         V, i = vertex_inputs.pop()
#         W, j = V.adjacent[i]
#         one_seg = []
#         while not W in G.vertices:
#             if W in G.edges:
#                 L = norm_label(W)
#                 A = end_label(W, j)
#                 B = end_label(W, j + 1)
#                 one_seg += [A, L]
#             W, j = W.flow(j)
#         one_seg.append(norm_label(W))
#         vertex_inputs.remove((W, j))
#         segments.append(one_seg)

#     return system_node_pos, other_node_pos, segments
        

    
# def draw_spatial_graph_in_3D(G, add_disks=False, z_height=20):
#     main_nodes, other_nodes, segs = position_spatial_graph_in_3D(G, z_height)
#     other_nodes.update(main_nodes)
#     nodes = other_nodes
#     scene = None
#     cmap = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
#             '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']

    
#     for i, (L, center) in enumerate(main_nodes.items()):
#         S = sphere(center, size=8, color=cmap[i % len(cmap)])
#         if scene is None:
#             scene = S
#         else:
#             scene += S

#         for i, one_seg in enumerate(segs):
#             scene += line3d([nodes[L] for L in one_seg],
#                         thickness=4, color=cmap[i % len(cmap)])

#     if add_disks:
#         for L, center in other_nodes.items():
#             if L.endswith('+'):
#                 x, y, z = center
#                 s = 6
#                 z = -0.5*z_height 
#                 scene += polygon3d([(x - s, y - s, z), (x + s, y - s, z),
#                                     (x + s, y + s, z), (x - s, y + s, z)], color='white')
    
#     return scene

  
        
                

# G = nx.check_planarity(nx.complete_graph(4))[1]
    




def get_coefficients_and_exponents(poly):

    """
    A helper function to extract the coefficients and exponents from a Yamada polynomial.

    The Yamada polynomial calculator was originally written with SageMath and the Laurent polynomial objects
    had explicit attributes for coefficients and exponents that you could directly query. However, switching
    to the cypari library to improve OS compatibility added a few complications, including that there is no native
    method to access the coefficients and exponents of Yamada polynomials. This function obtains them.
    """

    # Assumes all denominators are only A**n with no coefficient
    coefficients = poly.numerator().Vec()
    coeff_len = len(coefficients)

    exponents = []
    degree = poly.poldegree()

    for _ in range(coeff_len):
        exponents.append(degree)
        degree -= 1

    return coefficients, exponents





"""A module to computer to H polynomial of an Abstract Graph

Computation of the H polynomial is based on the following paper:

SYSTEMATIC ENUMERATION AND IDENTIFICATION OF UNIQUE SPATIAL TOPOLOGIES OF 3D SYSTEMS USING SPATIAL GRAPH REPRESENTATIONS
Peddada et al.
DOI: 10.1115/DETC2021-66900

P1. H(empty graph) = 1 and H(simple loop) = A + 1 + A^(-1)
P2. H(G_1 disjoint union G_2)      =  H(G_1) * H(G_2)
P3. H(G_1 union G_2)               = -H(G_1) * H(G_2)
P4. If G has a cut edge, then H(G) =  0
P5. Let e be a non-loop edge of graph G. Then H(G) = H(G/e) + H(G-e) where G/e is the graph obtained from G by
    contracting e to a point and G-e is G with e deleted.

"""

import networkx as nx
import collections
from cypari import pari


def has_cut_edge(abstract_graph):

    """
    Determines if the abstract graph has a cut edge. A cut edge is an edge that, if removed, would increase the number
    of disconnected sub-graphs in the graph.
    """

    g = nx.Graph(abstract_graph)

    for u, v in nx.bridges(g):
        if abstract_graph.number_of_edges(u, v) == 1:
            return True

    return False


def an_edge(graph):

    """
    A generator that returns the next edge in the graph.
    """

    return next(iter(graph.edges()))


def remove_valence_two_vertices(graph):

    """
    TODO Why this?
    """

    G = graph.copy()
    valence_two = {v for v, d in G.degree if d == 2}
    T = G.subgraph(valence_two)
    new_edges = []
    for comp in nx.connected_components(T):
        C = T.subgraph(comp)
        if C.number_of_nodes() == C.number_of_edges():
            x = next(iter(comp))
            y = x
        else:
            if len(comp) == 1:
                c = next(iter(comp))
                x, y = [e[1] for e in G.edges(c)]
            else:
                a, b = [c for c, d in C.degree if d == 1]
                x = [e[1] for e in G.edges(a) if e[1] not in comp][0]
                y = [e[1] for e in G.edges(b) if e[1] not in comp][0]
        new_edges.append((x, y))
    G.remove_nodes_from(valence_two)
    G.add_edges_from(new_edges)
    return G


def graph_hash(graph):

    """
    Returns a hash of the graph.

    The graph hash is used to cache the results of the H polynomial computation and avoid recomputing the H polynomial.
    """

    return nx.weisfeiler_lehman_graph_hash(graph, iterations=3)


H_poly_cache = collections.defaultdict(list)


def h_poly(g):

    """
    Computes the H polynomial of the abstract graph g.

    :param g: The abstract graph
    :return: The H polynomial of the abstract graph g
    """

    # Define the indeterminate A
    a = pari('A')

    # Validate the input format
    if not isinstance(g, nx.MultiGraph):
        g = nx.MultiGraph(g)


    # RECURSION BASE CONDITION
    # Abstract graph is empty: H Polynomial property P1: H(empty graph) = 1


    if g.number_of_nodes() == 0:
        return pari(1)


    # RECURSION LOGIC
    # Case 1: If all nodes of the graph are connected then recursively calculate the H polynomial of the graph.
    # Case 2: Else, the abstract graph contains subgraphs that are not connected. Recursively calculate the H polynomial
    # of each subgraph (see "if statement" above).


    # Recursion logic case 1: The abstract graph is connected.
    if nx.is_connected(g):

        # Check if the graph has already been computed
        its_hash = graph_hash(g)
        for H, poly in H_poly_cache[its_hash]:
            if nx.is_isomorphic(g, H):
                return poly
        g_for_cache = g.copy()

        # H Polynomial property P4
        # If G has a cut edge, then H(G) =  0
        if has_cut_edge(g):
            ans = pari(0)

        else:
            # TODO Why remove valence two vertices? Finish logic here...
            # Collapse valence two vertices (b/c do not affect calc...)
            g = remove_valence_two_vertices(g)

            loop_factor = pari(1)

            # P1 & P3

            loops = [e for e in g.edges() if e[0] == e[1]]

            for u, v in loops:
                g.remove_edge(u, v)
                loop_factor = -loop_factor * (a + 1 + a ** -1)

            # correcting for the difference between ...
            if g.number_of_nodes() == 1:
                ans = -loop_factor

            # TBD (check out the paper for formula, avoids P5 recursion?)
            elif g.number_of_nodes() == 2:

                b = -(a ** -1 + 2 + a)
                q = g.number_of_edges()

                h = ((b + 1) - (b + 1) ** q) / b

                ans = loop_factor * h

            # H Polynomial property P5.
            # H(G) = H(G/e) + H(G-e)
            else:
                e = an_edge(g)
                g_mod_e = nx.contracted_edge(g, e, self_loops=True)
                g_mod_e.remove_edge(e[0], e[0])
                g.remove_edge(*e)
                ans = (h_poly(g_mod_e) + h_poly(g)) * loop_factor

        H_poly_cache[its_hash].append((g_for_cache, ans))
        return ans

    # Recursion logic case 2: The abstract graph contains subgraphs that are not connected.
    else:

        # Must initialize ans as 1 for recursive multiplication.
        ans = pari(1)

        # H polynomial property P2
        # H(G_1 disjoint union G_2) = H(G_1) * H(G_2) where H(G_1) and H(G_2) are recursively computed.
        for vertices in nx.connected_components(g):
            S = g.subgraph(vertices).copy()
            h = h_poly(S)
            if h == 0:
                return pari(0)

            ans = ans * h

        return ans


from yamada.H_polynomial import h_poly, has_cut_edge, remove_valence_two_vertices

from yamada.spatial_graph_diagrams import SpatialGraphDiagram, Vertex, Edge, \
    Crossing, reverse_poly, normalize_yamada_polynomial

from yamada.spatial_graphs import SpatialGraph, LinearAlgebra

from yamada.spatial_graph_enumeration import extract_graph_from_json_file
